/* 
   Simple STEP/IGES File Viewer

   GL stuff based on wxWidgets "isosurf" example.
   
   T.W. 2013
*/

#include <wx/wx.h>

#include <wx/timer.h>
#include <wx/glcanvas.h>
#include <wx/math.h>
#include <wx/log.h>
#include <wx/cmdline.h>
#include <wx/wfstream.h>
#include <wx/zstream.h>
#include <wx/txtstrm.h>

#include "gl_window.h"
#include "s3d_model.h"

extern S3D_MODEL *LoadStepModel( const std::string& filename );
extern S3D_MODEL *LoadIgesModel( const std::string& filename );

// global options which can be set through command-line options
GLboolean g_doubleBuffer = GL_TRUE;
GLboolean g_smooth = GL_TRUE;
GLboolean g_lighting = GL_TRUE;

//---------------------------------------------------------------------------
// MyApp
//---------------------------------------------------------------------------

IMPLEMENT_APP(MyApp)

bool MyApp::OnInit()
{
    if ( !wxApp::OnInit() )
        return false;

    // Create the main frame window
    new MyFrame(NULL, wxT("STEP Viewer Sample"));

    return true;
}

void MyApp::OnInitCmdLine(wxCmdLineParser& parser)
{
/*    parser.AddSwitch("", "sb", "Do not use double buffering");
    parser.AddSwitch("", "db", "Use double buffering");
    parser.AddSwitch("", "va", "Use vertex arrays");

    wxApp::OnInitCmdLine(parser);*/
}

bool MyApp::OnCmdLineParsed(wxCmdLineParser& parser)
{
    /*if (parser.Found("sb"))
        g_doubleBuffer = GL_FALSE;
    else if (parser.Found("db"))
        g_doubleBuffer = GL_TRUE;

    
    return wxApp::OnCmdLineParsed(parser);*/
    return true;
}

//---------------------------------------------------------------------------
// MyFrame
//---------------------------------------------------------------------------

BEGIN_EVENT_TABLE(MyFrame, wxFrame)
    EVT_MENU(wxID_EXIT, MyFrame::OnExit)
    EVT_MENU(wxID_OPEN, MyFrame::OnMenuFileOpen)
END_EVENT_TABLE()

void MyFrame::OnMenuFileOpen( wxCommandEvent& WXUNUSED(event) )
{
    wxString filename = wxFileSelector(wxT("Choose STEP Model"), wxT(""), wxT(""), wxT(""),
        wxT("STEP File (*.stp)|*.stp|IGES File (*.igs)|*.igs"),
        wxFD_OPEN);

    if (!filename.IsEmpty())
    {
        m_canvas->LoadModel(filename);
        m_canvas->Refresh(false);
    }
}


MyFrame::MyFrame(wxFrame *frame, const wxString& title, const wxPoint& pos,
                 const wxSize& size, long style)
    : wxFrame(frame, wxID_ANY, title, pos, size, style),
      m_canvas(NULL)
{
    

    // Make a menubar
    wxMenu *fileMenu = new wxMenu;

    fileMenu->Append(wxID_OPEN, wxT("&Open..."));
    fileMenu->AppendSeparator();
    fileMenu->Append(wxID_EXIT, wxT("E&xit"));
    wxMenuBar *menuBar = new wxMenuBar;
    menuBar->Append(fileMenu, wxT("&File"));
    SetMenuBar(menuBar);


  // Make a TestGLCanvas

  // JACS
#ifdef __WXMSW__
    int *gl_attrib = NULL;
#else
    int gl_attrib[20] =
        { WX_GL_RGBA, WX_GL_MIN_RED, 1, WX_GL_MIN_GREEN, 1,
        WX_GL_MIN_BLUE, 1, WX_GL_DEPTH_SIZE, 1,
        WX_GL_DOUBLEBUFFER,
#  if defined(__WXMAC__) || defined(__WXCOCOA__)
        GL_NONE };
#  else
        None };
#  endif
#endif

    if (!g_doubleBuffer)
    {
        wxLogWarning("Disabling double buffering");

#ifdef __WXGTK__
        gl_attrib[9] = None;
#endif
        g_doubleBuffer = GL_FALSE;
    }

    m_canvas = new TestGLCanvas(this, wxID_ANY, gl_attrib);

    // Show the frame
    Show(true);
    Raise();

    m_canvas->InitGL();
}

MyFrame::~MyFrame()
{
    delete m_canvas;
}

// Intercept menu commands
void MyFrame::OnExit( wxCommandEvent& WXUNUSED(event) )
{
    // true is to force the frame to close
    Close(true);
}


//---------------------------------------------------------------------------
// TestGLCanvas
//---------------------------------------------------------------------------

BEGIN_EVENT_TABLE(TestGLCanvas, wxGLCanvas)
    EVT_SIZE(TestGLCanvas::OnSize)
    EVT_PAINT(TestGLCanvas::OnPaint)
    EVT_CHAR(TestGLCanvas::OnChar)
    EVT_MOUSE_EVENTS(TestGLCanvas::OnMouseEvent)
END_EVENT_TABLE()

TestGLCanvas::TestGLCanvas(wxWindow *parent,
                           wxWindowID id,
                           int* gl_attrib)
    : wxGLCanvas(parent, id, gl_attrib)
{
    m_xrot = 0;
    m_yrot = 0;
    m_zoom = 1.0;
    
    // Explicitly create a new rendering context instance for this canvas.
    m_glRC = new wxGLContext(this);
    m_model = NULL;
}

TestGLCanvas::~TestGLCanvas()
{
 //   if(m_model)
   //     delete m_model;
 
    delete m_glRC;
}

void TestGLCanvas::LoadModel(const wxString& filename)
{
    wxLocale l(wxLANGUAGE_ENGLISH);

    if(m_model)
        delete m_model;

    if(filename.EndsWith(".stp"))
        m_model = LoadStepModel ((const char *) filename.c_str() );
    else
        m_model = LoadIgesModel ((const char *) filename.c_str() );
    
    if(!m_model)
    {
        wxLogError(wxT("Can't load STEP file '%s'."), filename.c_str());   
        return;
    }
 
    m_model->BuildGLGeometry();   

    PostSizeEventToParent();
}

void TestGLCanvas::OnPaint( wxPaintEvent& WXUNUSED(event) )
{
    // This is a dummy, to avoid an endless succession of paint messages.
    // OnPaint handlers must always create a wxPaintDC.
    wxPaintDC dc(this);

    // This is normally only necessary if there is more than one wxGLCanvas
    // or more than one wxGLContext in the application.
    SetCurrent(*m_glRC);

    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
    glPushMatrix();
    glRotatef( m_yrot, 0.0f, 1.0f, 0.0f );
    glRotatef( m_xrot, 1.0f, 0.0f, 0.0f );
    glScalef( m_zoom, m_zoom, m_zoom );


    if(m_model)
        m_model->Render();


    glPopMatrix();
    glFlush(); // Not really necessary: buffer swapping below implies glFlush()

    SwapBuffers();
}

void TestGLCanvas::OnSize(wxSizeEvent& event)
{
    if ( !IsShownOnScreen() )
        return;
    // This is normally only necessary if there is more than one wxGLCanvas
    // or more than one wxGLContext in the application.
    SetCurrent(*m_glRC);

    // It's up to the application code to update the OpenGL viewport settings.
    // This is OK here only because there is only one canvas that uses the
    // context. See the cube sample for that case that multiple canvases are
    // made current with one context.
    glViewport(0, 0, event.GetSize().x, event.GetSize().y);
}

void TestGLCanvas::OnChar(wxKeyEvent& event)
{
    switch( event.GetKeyCode() )
    {
    case WXK_ESCAPE:
        wxTheApp->ExitMainLoop();
        return;

    case WXK_LEFT:
        m_yrot -= 15.0;
        break;

    case WXK_RIGHT:
        m_yrot += 15.0;
        break;

    case WXK_UP:
        m_xrot += 15.0;
        break;

    case WXK_DOWN:
        m_xrot -= 15.0;
        break;

    case 's': case 'S':
        g_smooth = !g_smooth;
        if (g_smooth)
            glShadeModel(GL_SMOOTH);
        else
            glShadeModel(GL_FLAT);
        break;

    case 'l': case 'L':
        g_lighting = !g_lighting;
        if (g_lighting)
            glEnable(GL_LIGHTING);
        else
            glDisable(GL_LIGHTING);
        break;

    default:
        event.Skip();
        return;
    }

    Refresh(false);
}

void TestGLCanvas::OnMouseEvent(wxMouseEvent& event)
{
    static int dragging = 0;
    static float last_x, last_y;
    int delta;
    // Allow default processing to happen, or else the canvas cannot gain focus
    // (for key events).
    event.Skip();

    delta = event.GetWheelRotation();
    if(delta < 0)
        m_zoom *= 0.9;
    else if (delta > 0)
        m_zoom *= 1.1;


    if(delta)
        Refresh(false);

    if (event.LeftIsDown())
    {
        if (!dragging)
        {
            dragging = 1;
        }
        else
        {
            m_yrot += (event.GetX() - last_x)*1.0;
            m_xrot += (event.GetY() - last_y)*1.0;
            Refresh(false);
        }
        last_x = event.GetX();
        last_y = event.GetY();
    } 
    else
    {
        dragging = 0;
    }
}

void TestGLCanvas::InitMaterials()
{
    static const GLfloat ambient[4] = {0.1f, 0.1f, 0.1f, 1.0f};
    static const GLfloat diffuse[4] = {1.0f, 1.0f, 1.0f, 1.0f};
    static const GLfloat position0[4] = {0.0f, 0.0f, 2000.0f, 0.0f};
    static const GLfloat position1[4] = {0.0f, 0.0f, -2000.0f, 0.0f};
    
    glLightfv(GL_LIGHT0, GL_AMBIENT, ambient);
    glLightfv(GL_LIGHT0, GL_DIFFUSE, diffuse);
    glLightfv(GL_LIGHT0, GL_POSITION, position0);
    glEnable(GL_LIGHT0);

    glLightfv(GL_LIGHT1, GL_AMBIENT, ambient);
    glLightfv(GL_LIGHT1, GL_DIFFUSE, diffuse);
    glLightfv(GL_LIGHT1, GL_POSITION, position1);
    glEnable(GL_LIGHT1);

    static const GLfloat lmodel_ambient[4] = {1.0f, 1.0f, 1.0f, 1.0f};
    static const GLfloat lmodel_twoside[1] = {GL_FALSE};

    glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
    glLightModelfv(GL_LIGHT_MODEL_TWO_SIDE, lmodel_twoside);

    glEnable(GL_LIGHTING);

}

void TestGLCanvas::InitGL()
{
    // Make the new context current (activate it for use) with this canvas.
    SetCurrent(*m_glRC);

    glClearColor(0.0f, 0.0f, 0.0f, 0.0f);

    glShadeModel(GL_SMOOTH);
    glEnable(GL_DEPTH_TEST);

    InitMaterials();

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glFrustum( -1.0, 1.0, -0.75, 0.5, 5.0, 2500.0 );

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glTranslatef( 0.0, 0.0, -100.0 );

    InitMaterials();
    
}

