#include "s3d_model.h"

#include <GL/gl.h>

S3D_MODEL::S3D_MODEL()
{
	m_glHandle = 0;
}

S3D_MODEL::~S3D_MODEL()
{
	if(m_glHandle)
		glDeleteLists(m_glHandle, 1);
}

void S3D_MODEL::BuildGLGeometry()
{
	m_glHandle = glGenLists(1);

	glNewList(m_glHandle, GL_COMPILE);
	
	printf("Faces: %d\n", m_faces.size());
	for(int i = 0; i < m_faces.size(); i++)
	{
		S3D_FACE *face = m_faces[i];
		face->BuildGLGeometry();
	}
	glEndList();
}	

void S3D_FACE::calcNormals()
{
	m_normals.clear();
	for(int i = 0; i < m_vertices.size(); i++)
		m_normals.push_back(S3D_VERTEX());

	for(int i = 0; i < m_tris.size(); i++)
	{
		const S3D_TRIANGLE& t = m_tris[i];
		const S3D_VERTEX& a = m_vertices[t.v[0]]; 
		const S3D_VERTEX& b = m_vertices[t.v[1]]; 
		const S3D_VERTEX& c = m_vertices[t.v[2]]; 

		S3D_VERTEX v1 (a.x - b.x, a.y - b.y, a.z - b.z);
		S3D_VERTEX v2 (c.x - b.x, c.y - b.y, c.z - b.z);
		
		S3D_VERTEX n (	v1.y * v2.z - v1.z * v2.y,
						v1.z * v2.x - v1.x * v2.z, 
						v1.x * v2.y - v1.y * v2.x);

		for(int j = 0; j < 3; j ++)
		{
			m_normals[t.v[j]].x += n.x; 
			m_normals[t.v[j]].y += n.y; 
			m_normals[t.v[j]].z += n.z; 
		}
	}

	for(int i = 0; i < m_normals.size(); i++)
	{
		S3D_VERTEX n = m_normals[i];
		
		float len = sqrt(n.x * n.x + n.y * n.y + n.z * n.z);

		if(len > 0.0)
		{
			n.x /= len;
			n.y /= len;
			n.z /= len;
		}
		m_normals[i] = n;
	}
}

void S3D_MODEL::AddFace ( S3D_FACE *face )
{
	m_faces.push_back(face);
}

S3D_FACE::~S3D_FACE()
{

}

void S3D_FACE::BuildGLGeometry()
{
	S3D_COLOR color (0.0, 1.0, 0.0);
	
	if(m_material)
		color = m_material->diffuse;

	calcNormals();

	glBegin(GL_TRIANGLES);

	float tmp[4];

	tmp[0] = m_material->shininess;
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, tmp);
	tmp[0] = m_material->specular.r;
	tmp[1] = m_material->specular.g;
	tmp[2] = m_material->specular.b;
	tmp[3] = m_material->specular.a;
	
    glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, tmp);
    tmp[0] = m_material->diffuse.r;
	tmp[1] = m_material->diffuse.g;
	tmp[2] = m_material->diffuse.b;
	tmp[3] = m_material->diffuse.a;
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, tmp);
	
	for(int i = 0; i< m_tris.size(); i++)
	{
		const S3D_TRIANGLE& t = m_tris[i];
		
		//glColor4f(color.r, color.g, color.b, color.a);
		glNormal3f( m_normals[t.v[0]].x, m_normals[t.v[0]].y, m_normals[t.v[0]].z );
		glVertex3f( m_vertices[t.v[0]].x, m_vertices[t.v[0]].y, m_vertices[t.v[0]].z );
		glNormal3f( m_normals[t.v[1]].x, m_normals[t.v[1]].y, m_normals[t.v[1]].z );
		glVertex3f( m_vertices[t.v[1]].x, m_vertices[t.v[1]].y, m_vertices[t.v[1]].z );
		glNormal3f( m_normals[t.v[2]].x, m_normals[t.v[2]].y, m_normals[t.v[2]].z );
		glVertex3f( m_vertices[t.v[2]].x, m_vertices[t.v[2]].y, m_vertices[t.v[2]].z );
	}
	glEnd();
}

void S3D_MODEL::Render()
{
	glCallList(m_glHandle);
}

S3D_MATERIAL::S3D_MATERIAL()
{
	specular = S3D_COLOR(1.0, 1.0, 1.0, 1.0);
	shininess = 60.0;
}